#if !defined(AFX_UTP__H__INCLUDED_)
#define AFX_UTP__H__INCLUDED_



//Copyright (c) 2010 BitTorrent, Inc.
//
//Permission is hereby granted, free of charge, to any person obtaining a copy
//of this software and associated documentation files (the "Software"), to deal
//in the Software without restriction, including without limitation the rights
//to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//copies of the Software, and to permit persons to whom the Software is
//furnished to do so, subject to the following conditions:
//
//The above copyright notice and this permission notice shall be included in
//all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//THE SOFTWARE.


#define cstr		char*
#define uint32_big	unsigned int
#define uint16_big	unsigned short
#define byte		unsigned char



#define PACKET_SIZE 300


#pragma pack(push,1)

struct PacketFormat
{
	uint32_big		connid;		// connection ID
	uint32_big		tv_sec;
	uint32_big		tv_usec;
	uint32_big		reply_micro;
	byte			windowsize;	// receive window size in PACKET_SIZE chunks
	byte			ext;		// Type of the first extension header
	byte			flags;		// Flags
	uint16_big		seq_nr;		// Sequence number
	uint16_big		ack_nr;		// Acknowledgment number
};


struct PacketFormatV1
{
	byte			version:4;	// protocol version
	byte			type:4;		// type (formerly flags)
	byte			ext;		// Type of the first extension header
	uint16_big		connid;		// connection ID
	uint32_big		tv_usec;
	uint32_big		reply_micro;
	uint32_big		windowsize;	// receive window size in bytes
	uint16_big		seq_nr;		// Sequence number
	uint16_big		ack_nr;		// Acknowledgment number
};


#pragma pack(pop)



enum {
	ST_DATA		= 0,	// Data packet.
	ST_FIN		= 1,	// Finalize the connection. This is the last packet.
	ST_STATE	= 2,	// State packet. Used to transmit an ACK with no data.
	ST_RESET	= 3,	// Terminate connection forcefully.
	ST_SYN		= 4,	// Connect SYN
	ST_NUM_STATES,		// used for bounds checking
};


static const cstr flagnames[]= {"ST_DATA","ST_FIN","ST_STATE","ST_RESET","ST_SYN"};



/*
 * Copyright (c) 2010 Rozhuk Ivan <rozhuk.im@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */



#pragma pack(push,1)

// non standart: removed PacketFormat[V1]
struct PFAckUni
{
	byte	ext_next;
	byte	ext_len;
	byte	acks[4];
};

struct PFExtUni
{
	byte	ext_next;
	byte	ext_len;
	byte	extensions[8];
};


struct PFAckExtUni
{
	byte	ext_next;
	byte	ext_len;
};


#pragma pack(pop)




// 0 none/end
// 1 Selective Acknowledgment
// 2 extension bits
#define MAX_EXT_NUM		3




#ifndef NTOHS

__inline u_short USHORT_FLIP(u_short usIn)
{	
return(((usIn<<8) | (usIn>>8)));
}
#define NTOHS(in)   USHORT_FLIP(in)
#define HTONS(in)   USHORT_FLIP(in)

#endif



#ifndef NTOHL

__inline u_long ULONG_FLIP(u_long ulIn)
{
return((((ulIn<<8) & 0x00ff0000) | (ulIn<<24) | ((ulIn>>8) & 0x0000ff00) | (ulIn>>24)));
}

#define NTOHL(in) ULONG_FLIP(in)
#define HTONL(in) ULONG_FLIP(in)

#endif


//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////Packet vars/////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

__inline unsigned int GetConnIDFromPacketuTP(int iVersion,unsigned char *pData)
{// connection ID
	unsigned int uiRet;

	switch(iVersion){
	case 0:
		uiRet=NTOHL(((PacketFormat*)pData)->connid);
		break;
	case 1:
		uiRet=NTOHS(((PacketFormatV1*)pData)->connid);
		break;
	default:
		uiRet=0;
	}
return(uiRet);
}

__inline unsigned int GetTVUsecFromPacketuTP(int iVersion,unsigned char *pData)
{//
	unsigned int uiRet;

	switch(iVersion){
	case 0:
		uiRet=((NTOHL(((PacketFormat*)pData)->tv_sec)*1000000) + NTOHL(((PacketFormat*)pData)->tv_usec));
		break;
	case 1:
		uiRet=NTOHL(((PacketFormatV1*)pData)->tv_usec);
		break;
	default:
		uiRet=0;
	}
return(uiRet);
}


__inline unsigned int GetReplyMicroFromPacketuTP(int iVersion,unsigned char *pData)
{//
	unsigned int uiRet;

	switch(iVersion){
	case 0:
		uiRet=NTOHL(((PacketFormat*)pData)->reply_micro);
		uiRet=((uiRet==INT_MAX)? 0:uiRet);
		break;
	case 1:
		uiRet=NTOHL(((PacketFormatV1*)pData)->reply_micro);
		uiRet=((uiRet==INT_MAX)? 0:uiRet);// ? - violete specification
		break;
	default:
		uiRet=0;
	}
return(uiRet);
}

__inline unsigned int GetWindowSizeFromPacketuTP(int iVersion,unsigned char *pData)
{// receive window size
	unsigned int uiRet;

	switch(iVersion){
	case 0:
		uiRet=(((PacketFormat*)pData)->windowsize * PACKET_SIZE);// receive window size in PACKET_SIZE chunks
		break;
	case 1:
		uiRet=(((PacketFormatV1*)pData)->windowsize);// receive window size in bytes
		break;
	default:
		uiRet=0;
	}
return(uiRet);
}


__inline unsigned short GetSeqNrFromPacketuTP(int iVersion,unsigned char *pData)
{// Sequence number
	unsigned short usRet;

	switch(iVersion){
	case 0:
		usRet=NTOHS(((PacketFormat*)pData)->seq_nr);
		break;
	case 1:
		usRet=NTOHS(((PacketFormatV1*)pData)->seq_nr);
		break;
	default:
		usRet=0;
	}
return(usRet);
}

__inline unsigned short GetAckNrFromPacketuTP(int iVersion,unsigned char *pData)
{// Acknowledgment number
	unsigned short usRet;

	switch(iVersion){
	case 0:
		usRet=NTOHS(((PacketFormat*)pData)->ack_nr);
		break;
	case 1:
		usRet=NTOHS(((PacketFormatV1*)pData)->ack_nr);
		break;
	default:
		usRet=0;
	}
return(usRet);
}


__inline unsigned char GetFlagsTypeFromPacketuTP(int iVersion,unsigned char *pData)
{// type (formerly flags) = ST_*
	unsigned char btRet;

	switch(iVersion){
	case 0:
		btRet=((PacketFormat*)pData)->flags;
		break;
	case 1:
		btRet=((PacketFormatV1*)pData)->type;
		break;
	default:
		btRet=0;
	}
return(btRet);
}


__inline unsigned char GetExtFromPacketuTP(int iVersion,unsigned char *pData)
{// Type of the first extension header
	unsigned char btRet;

	switch(iVersion){
	case 0:
		btRet=((PacketFormat*)pData)->ext;
		break;
	case 1:
		btRet=((PacketFormatV1*)pData)->ext;
		break;
	default:
		btRet=0;
	}
return(btRet);
}
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////Packet process//////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
__inline int IsValidPacketuTPExt(byte btExt,unsigned char *pData,size_t dwDataSize)
{
	int iRet=0;

	if (btExt)
	{
		if (btExt<MAX_EXT_NUM)
		{// Verify that the packet is valid. %)
			PFAckExtUni *pPFAE;

			iRet=1;
			pPFAE=(PFAckExtUni*)pData;
			do
			{
				if (btExt<MAX_EXT_NUM && dwDataSize>=pPFAE->ext_len && !(btExt==2 && pPFAE->ext_len!=8))
				{
					btExt=pPFAE->ext_next;
					dwDataSize-=pPFAE->ext_len;
					pPFAE=(PFAckExtUni*)(((unsigned char*)pPFAE)+pPFAE->ext_len);
				}else{
					iRet=0;
					break;
				}
			}while(btExt);
		}
	}else{
		iRet=1;
	}
return(iRet);
}


__inline int IsPacketuTPv0(unsigned char *pData,size_t dwDataSize)
{
	int iRet=0;

	if (dwDataSize>=sizeof(PacketFormat))
	{// size OK
		PacketFormat *pPF=(PacketFormat*)pData;

		if (pPF->flags<ST_NUM_STATES && pPF->ext<3)
		{
			iRet=IsValidPacketuTPExt(pPF->ext,(pData+sizeof(PacketFormat)),(dwDataSize-sizeof(PacketFormat)));
		}
	}
return(iRet);
}


__inline int IsPacketuTPv1(unsigned char *pData,size_t dwDataSize)
{
	int iRet=0;

	if (dwDataSize>=sizeof(PacketFormatV1))
	{// size OK
		PacketFormatV1 *pPFv1=(PacketFormatV1*)pData;

		if (pPFv1->version==1 && pPFv1->type<ST_NUM_STATES && pPFv1->ext<MAX_EXT_NUM)
		{
			iRet=IsValidPacketuTPExt(pPFv1->ext,(pData+sizeof(PacketFormatV1)),(dwDataSize-sizeof(PacketFormatV1)));
		}
	}
return(iRet);
}

__inline int GetPacketuTPVersion(unsigned char *pData,size_t dwDataSize)
{
	int iRet;

	iRet=-1;
	if (IsPacketuTPv1(pData,dwDataSize))
	{
		iRet=1;
	}else
	if (IsPacketuTPv0(pData,dwDataSize))
	{
		iRet=0;
	}
return(iRet);
}





__inline size_t GenRSTPacketuTP(int iVersion,unsigned int uiConnIDSend,unsigned short usAckNr,unsigned short usSeqNr,unsigned char *pBuff,size_t dwBuffSize)
{
	size_t dwRet;

	switch(iVersion){
	case 0:
		if (dwBuffSize>=sizeof(PacketFormat))
		{
			PacketFormat *pPF=(PacketFormat*)pBuff;

			bzero(pBuff,sizeof(PacketFormat));
			dwRet=sizeof(PacketFormat);

			pPF->connid=HTONL(uiConnIDSend);
			pPF->ack_nr=HTONS(usAckNr);
			pPF->seq_nr=HTONS(usSeqNr);
			pPF->flags=ST_RESET;
			//pPF->ext=0;
			//pPF->windowsize=0;
		}
		break;
	case 1:
		if (dwBuffSize>=sizeof(PacketFormatV1))
		{
			PacketFormatV1 *pPFv1=(PacketFormatV1*)pBuff;

			bzero(pBuff,sizeof(PacketFormatV1));
			dwRet=sizeof(PacketFormatV1);

			pPFv1->version=1;
			pPFv1->type=ST_RESET;
			//pPFv1->ext=0;
			pPFv1->connid=HTONS(uiConnIDSend);
			pPFv1->ack_nr=HTONS(usAckNr);
			pPFv1->seq_nr=HTONS(usSeqNr);
			//pPFv1->windowsize=0;
		}
		break;
	default:
		dwRet=0;
	}
return(dwRet);
}







#endif // AFX_UTP__H__INCLUDED_

