/*
 * Copyright (c) 2010 Rozhuk Ivan <rozhuk.im@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include <sys/param.h>
#include <sys/cpuset.h>
#include <sys/stat.h>
#include <errno.h>
#include <iostream>
#include <unistd.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>



#include <default.h>
#include <uTP.h>


#define THREAD_BUFF_SIZE	(256*1024)




struct PROG_SETTINGS
{
	int				iSktDivert;
	unsigned short	usBindPort;
	char			*pPidFileName;
	int				iCPUID;
	uid_t			pwUID;// user uid
	gid_t			pwGID;// user gid
};


PROG_SETTINGS gpsProgSettings;




int					GetOptions				(int argc, char **argv);
void				SigInstall				();
static void			SigHandler				(int iSigNum);
void				MakeDaemon				();
int					WritePID				(char *pPIDFileName);
int					BindThreadToCPU			(int iCPUID);
int					ProcessPacket			(int iSocket,struct sockaddr_in *psiIFAddr,socklen_t slIFAddrLen,unsigned char *pBuff,size_t dwDataSize,size_t dwBuffSize);




int main(int argc,char *argv[])
{
	int iRet=0;
	
	bzero(&gpsProgSettings,sizeof(gpsProgSettings));

	gpsProgSettings.iCPUID=-1;
	if (GetOptions(argc,argv)==0)
	{
		int iSktDivert,iRunning;

		fprintf(stdout,"%s %s: started\n",PACKAGE,VERSION);

		iSktDivert=socket(PF_INET,SOCK_RAW,IPPROTO_DIVERT);
		if (iSktDivert!=-1)
		{// raw divert socket created
			struct sockaddr_in siAddr;

			siAddr.sin_family=PF_INET;
			siAddr.sin_port=HTONS(gpsProgSettings.usBindPort);
			siAddr.sin_addr.s_addr=INADDR_ANY;

			iRet=bind(iSktDivert,(struct sockaddr*)&siAddr,sizeof(siAddr));
			if (iRet!=-1) 
			{// socket binded to specified port
				size_t dwReceived;
				socklen_t slIFAddrLen;
				unsigned char btBuff[THREAD_BUFF_SIZE];

				SigInstall();

				gpsProgSettings.iSktDivert=iSktDivert;
				BindThreadToCPU(gpsProgSettings.iCPUID);
				if (gpsProgSettings.pPidFileName) WritePID(gpsProgSettings.pPidFileName);// Store pid to file
				if (gpsProgSettings.pwGID) if (setgid(gpsProgSettings.pwGID)!=0) fprintf(stderr,"setgid() failed: %i %s\n",errno,strerror(errno));
				if (gpsProgSettings.pwUID) if (setuid(gpsProgSettings.pwUID)!=0) fprintf(stderr,"setuid() failed: %i %s\n",errno,strerror(errno));


				// read data in
				while(1)
				{
					slIFAddrLen=sizeof(siAddr);
					dwReceived=recvfrom(iSktDivert,btBuff,THREAD_BUFF_SIZE,0,(struct sockaddr*)&siAddr,&slIFAddrLen);
					if (dwReceived!=(size_t)-1)
					{// received some packet
						ProcessPacket(iSktDivert,&siAddr,slIFAddrLen,btBuff,dwReceived,THREAD_BUFF_SIZE);
					}else{// exit on error
						fprintf(stderr,"%s: recvfrom error: %i %s\n",PACKAGE,errno,strerror(errno));
						break;
					}
				}

				if (gpsProgSettings.pPidFileName) unlink(gpsProgSettings.pPidFileName);// Remove pid file
			}else{
				fprintf(stderr,"%s: cant bind() divert socket: %i %s\n",PACKAGE,errno,strerror(errno));
			}
			close(iSktDivert);
		}else{
			fprintf(stderr,"%s: socket() cant create divert socket: %i %s\n",PACKAGE,errno,strerror(errno));
		}

		fprintf(stdout,"%s: exiting\n",argv[0]);
	}else{
		fprintf(stderr, "%s %s -- (c) Rozhuk Ivan <rozhuk.im@gmail.com>\n",PACKAGE,VERSION);
		fprintf(stderr, "   BSD licence.  Website: http://rozhuk.org.ru\n");
		fprintf(stderr,
		       "usage: %s [-d] [-b port] [-c CPUnum] [-p port]\n"
                       "       [-p PID file] [-u uid|usr -g gid|grp]\n"
		       " -h           usage (this screen)\n"
		       " -d           become daemon\n"
		       " -b port      bind to port\n",PACKAGE);
		fprintf(stderr,
		       " -c CPUnum    bind thread to CPU (default no bind)\n"
		       " -p PID file  file name to store PID\n"
		       " -u uid|user  change uid\n"
		       " -g gid|group change gid\n");
	}
return(iRet);
}


int GetOptions(int argc, char **argv)
{
	int iRet=0;
	
	if (argc>1)
	{
		unsigned char *pCur;
		size_t i;

		iRet=0;
		for (i=1;i<argc;i++)
		{
			pCur=(unsigned char*)argv[i];

			if (*pCur++!='-')
			{
				fprintf(stderr,"invalid option: \"%s\"\n", argv[i]);
				iRet=1;
			}

			while(*pCur && iRet==0)
			{
				switch(*pCur++){
				case '?':case 'h':case 'H':
					iRet=1;
					break;
				case 'd':case 'D':
					MakeDaemon();
					break;
				case 'b':case 'B':
					if (*pCur)
					{
						gpsProgSettings.usBindPort=atol((char*)pCur);
					}else
					if (argv[++i])
					{
						gpsProgSettings.usBindPort=atol(argv[i]);
					}else{
						fprintf(stderr,"option \"-b\" requires bind port number\n");
						iRet=1;
					}
					break;
				case 'c':case 'C':
					if (*pCur)
					{
						gpsProgSettings.iCPUID=atol((char*)pCur);
					}else
					if (argv[++i])
					{
						gpsProgSettings.iCPUID=atol(argv[i]);
					}else{
						fprintf(stderr,"option \"-c\" requires CPU number\n");
						iRet=1;
					}
					break;
				case 'p':case 'P':
					if (*pCur)
					{
						gpsProgSettings.pPidFileName=(char*)pCur;
					}else
					if (argv[++i])
					{
						gpsProgSettings.pPidFileName=argv[i];
					}else{
						fprintf(stderr,"option \"-p\" requires pid file name\n");
						iRet=1;
					}
					break;
				case 'u':case 'U':
					{
						char *pUID;

						if (*pCur)
						{
							pUID=(char*)pCur;
						}else
						if (argv[++i])
						{
							pUID=argv[i];
						}else{
							fprintf(stderr,"option \"-u\" requires UID\n");
							pUID=NULL;
							iRet=1;
						}

						if (pUID)
						{
							struct passwd *pwd;

							pwd=getpwnam(pUID);
							if (pwd)
							{
								gpsProgSettings.pwUID=pwd->pw_uid;
							}else{
								fprintf(stderr,"option \"-u\" requires UID, UID %s not found\n",pUID);
							}
						}
					}
					break;
				case 'g':case 'G':
					{
						char *pGID;

						if (*pCur)
						{
							pGID=(char*)pCur;
						}else
						if (argv[++i])
						{
							pGID=argv[i];
						}else{
							fprintf(stderr,"option \"-g\" requires GID\n");
							pGID=NULL;
							iRet=1;
						}

						if (pGID)
						{
							struct group *grp;

							grp=getgrnam(pGID);
							if (grp)
							{
								gpsProgSettings.pwGID=grp->gr_gid;
							}else{
								fprintf(stderr,"option \"-g\" requires GID, GID %s not found\n",pGID);
							}
						}
					}
					break;
				default:
					fprintf(stderr,"invalid option: \"%c\"\n",*(pCur-1));
					iRet=1;
				}
			}
		}
	}else{
		iRet=1;
	}
return(iRet);
}


void SigInstall()
{
	signal(SIGINT,	SigHandler);
	signal(SIGTERM,	SigHandler);
	signal(SIGKILL,	SigHandler);
	signal(SIGHUP,	SigHandler);
	signal(SIGUSR1,	SigHandler);
	signal(SIGUSR2,	SigHandler);
	signal(SIGPIPE,	SIG_IGN);
}

static void SigHandler(int iSigNum)
{
	switch(iSigNum){
	case SIGINT:
	case SIGTERM:
	case SIGKILL:
		close(gpsProgSettings.iSktDivert);
		break;
	case SIGHUP:
	case SIGUSR1:
	case SIGUSR2:
		break;
	}
}


void MakeDaemon()
{
    pid_t pdDaemon;

	pdDaemon=fork();
    if (pdDaemon!=-1)
	{
		if (pdDaemon)
		{//Exit Parent
			exit(0);
		}else{//Child
			setsid();
			setpgid(getpid(),0);
			chdir("/tmp/");
			umask(077);

			//Close stdin/stdout/stderr
			close(0);
			close(1);
			close(2);
		}
	}else{
		fprintf(stderr,"MakeDaemon: fork() failed: %i %s\n",errno,strerror(errno));
	}
}


int WritePID(char *pPIDFileName)
{
	int iRet;
	int iFileHandle;
	
	iFileHandle=open(pPIDFileName,(O_RDWR|O_CREAT|O_TRUNC));
	if (iFileHandle==-1)
	{
		fprintf(stderr,"WritePID: fopen() failed: %i %s\n",errno,strerror(errno));
		iRet=0;
	}else{
		char szPid[64];
		size_t dwPidSize;
		
		dwPidSize=snprintf(szPid,sizeof(szPid),"%ld",getpid());
		write(iFileHandle,szPid,dwPidSize);
		fchmod(iFileHandle,(S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH));
		close(iFileHandle);
		iRet=1;
	}
return(iRet);
}



int BindThreadToCPU(int iCPUID)
{
	int bRet=0;

	if (iCPUID!=-1)
	{
		cpuset_t csCPUSet;

		// bind this thread to a single cpu
		CPU_ZERO(&csCPUSet);
		CPU_SET(iCPUID,&csCPUSet);
		if (cpuset_setaffinity(CPU_LEVEL_WHICH,CPU_WHICH_TID,-1,sizeof(csCPUSet),&csCPUSet)==-1)
		{
			fprintf(stderr,"BindThreadToCPU: can't bind thread to CPU %s\n",strerror(errno));
		}else{
			bRet=1;
		}
	}
return(bRet);
}



int ProcessPacket(int iSocket,struct sockaddr_in *psiIFAddr,socklen_t slIFAddrLen,unsigned char *pBuff,size_t dwDataSize,size_t dwBuffSize)
{
	int iRet=0;

	if (((struct ip*)pBuff)->ip_p==IPPROTO_UDP)
	{
		int iVersion;
		size_t dwUDPDataSize;
		struct ip *pIPHdr;
		struct udphdr *pUDPHdr;
		unsigned char *pUDPData;

		pIPHdr=(struct ip*)pBuff;
		pUDPHdr=(struct udphdr*)(pIPHdr+1);
		pUDPData=(unsigned char*)(pUDPHdr+1);
		dwUDPDataSize=(NTOHS(pUDPHdr->uh_ulen)-sizeof(struct udphdr));

		iVersion=GetPacketuTPVersion(pUDPData,dwUDPDataSize);
		if (iVersion!=-1)
		{
			unsigned int uiConnIDSend,uiFlagsType;
			unsigned short usAckNr,usSeqNr;

			uiFlagsType=GetFlagsTypeFromPacketuTP(iVersion,pUDPData);
			
			//time_t tv_usec=GetTVUsecFromPacketuTP(iVersion,pUDPData);//,ctime(&tv_usec)
			//printf("\nAddress: %s",inet_ntoa(pIPHdr->ip_src));
			//printf(" -> %s",inet_ntoa(pIPHdr->ip_dst));
			//printf(" via %s, Protocol number: %hhu, checksum=%hu, ip_len=%hu, BuffSize=%lu\n",inet_ntoa(psiIFAddr->sin_addr),pIPHdr->ip_p,NTOHS(pIPHdr->ip_sum),NTOHS(pIPHdr->ip_len),dwDataSize);
			//printf("UDP from port %hu to %hu, checksum=%hu, len=%hu, UDPDataSize=%lu or =%lu\n",HTONS(pUDPHdr->uh_sport),HTONS(pUDPHdr->uh_dport),NTOHS(pUDPHdr->uh_sum),NTOHS(pUDPHdr->uh_ulen),dwUDPDataSize,(dwDataSize-(pUDPData-pBuff)));
			//printf("uTP Version=%i, TypeFlag=%s, ext=%hhu, conn_id_send=%lu, ack_nr=%hu, seq_nr=%hu, windowsize=%lu, reply_micro=%lu, tv_usec=%lu\n",iVersion,flagnames[uiFlagsType],GetExtFromPacketuTP(iVersion,pUDPData),uiConnIDSend,GetAckNrFromPacketuTP(iVersion,pUDPData),GetSeqNrFromPacketuTP(iVersion,pUDPData),GetWindowSizeFromPacketuTP(iVersion,pUDPData),GetReplyMicroFromPacketuTP(iVersion,pUDPData),tv_usec);
			
			//if (uiFlagsType==ST_STATE || uiFlagsType==ST_SYN)
			if (uiFlagsType!=ST_FIN && uiFlagsType!=ST_RESET)
			{// send reset with all packet types, except reset, fin
				uiConnIDSend=GetConnIDFromPacketuTP(iVersion,pUDPData);
				usAckNr=GetAckNrFromPacketuTP(iVersion,pUDPData);
				usSeqNr=(usAckNr+1);

				dwUDPDataSize=GenRSTPacketuTP(iVersion,uiConnIDSend,usAckNr,usSeqNr,pUDPData,(dwBuffSize-(sizeof(struct ip)+sizeof(struct udphdr))));
				if (dwUDPDataSize)
				{
					pUDPHdr->uh_ulen=HTONS((dwUDPDataSize+sizeof(struct udphdr)));
					pUDPHdr->uh_sum=0;// auto recalc sum UDP
					pIPHdr->ip_len=HTONS((dwUDPDataSize+(sizeof(struct ip)+sizeof(struct udphdr))));
					pIPHdr->ip_sum=0;// auto recalc sum IP

					psiIFAddr->sin_addr.s_addr=INADDR_ANY;
					iRet=sendto(iSocket,pBuff,dwDataSize,MSG_NOSIGNAL,(struct sockaddr*)psiIFAddr,slIFAddrLen);
					if (iRet==-1)
					{
						fprintf(stderr,"ProcessPacket: write error %i %s\n",errno,strerror(errno));
					}
				}
			}
		}
	}

return(iRet);
}



